from __future__ import annotations

from datetime import datetime, timezone
from typing import Any

from pgqueuer.models import JobId, TracebackRecord


def test_from_exception() -> None:
    job_id: JobId = JobId(123)
    additional: dict[str, str] = {"detail": "Error occurred"}

    try:
        raise ValueError("Testing error")
    except Exception as exc:
        record: TracebackRecord = TracebackRecord.from_exception(exc, job_id, additional)

    assert record.job_id == job_id
    assert isinstance(record.job_id, int)
    assert record.exception_type == "ValueError"
    assert "Testing error" in record.exception_message
    assert isinstance(record.timestamp, datetime)
    assert len(record.traceback) > 0
    assert record.additional_context == additional


def test_timestamp() -> None:
    job_id: JobId = JobId(456)

    try:
        raise RuntimeError("Another test error")
    except Exception as exc:
        record: TracebackRecord = TracebackRecord.from_exception(exc, job_id)
    now: datetime = datetime.now(timezone.utc)

    assert record.timestamp <= now
    diff: float = (now - record.timestamp).total_seconds()
    assert diff < 5  # Ensure the recorded timestamp is recent


def test_model_dump() -> None:
    job_id: JobId = JobId(789)
    additional: dict[str, str] = {"key": "value"}
    try:
        raise KeyError("Missing key")
    except Exception as exc:
        record: TracebackRecord = TracebackRecord.from_exception(exc, job_id, additional)
    record_dict: dict[str, Any] = record.model_dump()
    expected_keys: set[str] = {
        "job_id",
        "timestamp",
        "exception_type",
        "exception_message",
        "traceback",
        "additional_context",
    }
    assert set(record_dict.keys()) == expected_keys
    assert record_dict["job_id"] == job_id
    assert isinstance(record_dict["timestamp"], datetime)
    assert record_dict["additional_context"] == additional


def test_default_additional_context() -> None:
    job_id: JobId = JobId(1)
    try:
        raise Exception("Generic error")
    except Exception as exc:
        record: TracebackRecord = TracebackRecord.from_exception(exc, job_id)
    assert record.additional_context is None



from __future__ import annotations

import asyncio
from contextlib import suppress
from datetime import datetime, timedelta

import asyncpg
import icecream

try:
    from uvloop import run as asyncio_run
except ImportError:
    from asyncio import run as asyncio_run  # type: ignore[assignment]

from pgqueuer.db import AsyncpgDriver, dsn
from pgqueuer.models import Job
from pgqueuer.qb import DBSettings
from pgqueuer.qm import QueueManager
from pgqueuer.queries import Queries


async def probe(driver: AsyncpgDriver) -> None:
    while True:
        jobs = await driver.fetch(
            f"SELECT COUNT(*), entrypoint, status FROM {DBSettings().queue_table} "
            "GROUP BY (entrypoint, status) ORDER BY entrypoint, status"
        )
        icecream.ic(jobs)
        await asyncio.sleep(1)


async def producer(driver: AsyncpgDriver) -> None:
    queries = Queries(driver)
    await queries.clear_queue()
    await queries.clear_statistics_log()
    for i in range(100):
        for j in range(1, 10):
            await queries.enqueue(f"fetch_{j}", str(i).encode())

    for i in range(100):
        await queries.enqueue("fetch_0", str(i).encode())


async def consumer(driver: AsyncpgDriver) -> None:
    qm = QueueManager(driver)

    for i in range(1, 10):

        @qm.entrypoint(f"fetch_{i}", retry_timer=timedelta(minutes=2), concurrency_limit=1)
        async def process_message(job: Job) -> None:
            await asyncio.sleep(3)

    @qm.entrypoint("fetch_0", retry_timer=timedelta(minutes=3), concurrency_limit=1)
    async def fetch2(job: Job) -> None:
        assert job.payload
        print("fetch_0 start", job.payload.decode(), datetime.now().astimezone())
        await asyncio.sleep(10)

    print("startup", datetime.now().astimezone())
    await qm.run(batch_size=1, dequeue_timeout=timedelta(minutes=1))


async def main() -> None:
    connection = await asyncpg.connect(dsn())
    driver = AsyncpgDriver(connection)

    await producer(driver)
    await asyncio.gather(consumer(driver), probe(driver))


if __name__ == "__main__":
    with suppress(asyncio.CancelledError, KeyboardInterrupt):
        asyncio_run(main())



# /// script
# requires-python = ">=3.12"
# dependencies = [
#     "plotly",
#     "pydantic",
#     "httpx",
# ]
# ///

from __future__ import annotations

import os
import re
from collections import Counter, defaultdict
from datetime import datetime, timedelta

import httpx
import plotly.graph_objects as go
import plotly.subplots as sp
from pydantic import BaseModel, Field


def pepy_api_key() -> str:
    """Return PEPY API key from env var PEPY_API_KEY."""
    if key := os.environ.get("PEPY_API_KEY"):
        return key
    raise RuntimeError("Missing env: `PEPY_API_KEY`")


class PackageStats(BaseModel):
    """Represent package download statistics from PEPY."""

    total_downloads: int = Field(..., description="Total downloads of all versions")
    id: str = Field(..., description="Package name")
    versions: list[str] = Field(..., description="List of available versions")
    downloads: dict[datetime, dict[str, int]] = Field(
        ..., description="Daily download counts per version"
    )

    class Config:
        extra = "forbid"


def fetch() -> PackageStats:
    """Fetch and return package stats from PEPY."""
    return PackageStats.model_validate_json(
        httpx.get(
            "https://api.pepy.tech/api/v2/projects/pgqueuer",
            headers={"X-API-Key": pepy_api_key()},
        ).content
    )


def plot_downloads(data: PackageStats) -> None:
    """Create and display a Plotly figure of the package download stats."""
    downloads: defaultdict[datetime, defaultdict[str, int]] = defaultdict(lambda: defaultdict(int))
    for date, vers_counts in data.downloads.items():
        for version, count in vers_counts.items():
            if mv := re.match(r"^\d+\.\d+", version):
                downloads[date][mv.group(0)] += count

    accumulated = defaultdict[datetime, int](int)
    for date in sorted(downloads.keys()):
        accumulated[date] = sum(downloads[date].values()) + accumulated.get(
            date - timedelta(days=1), 0
        )

    totals = Counter[str]()
    for vers_counts in downloads.values():
        for version, count in vers_counts.items():
            totals[version] += count

    dates = sorted(downloads.keys())
    versions = sorted({v for vc in downloads.values() for v in vc}, key=lambda x: (len(x), x))

    fig = sp.make_subplots(
        rows=2,
        cols=2,
        specs=[[{"type": "xy"}, {"type": "xy"}], [{"type": "xy", "colspan": 2}, None]],
        subplot_titles=[
            "Downloads by Version",
            "Accumulated Downloads",
            "Version Distribution (Bar)",
        ],
    )

    for version in versions:
        fig.add_trace(
            go.Scatter(
                x=dates,
                y=[downloads[d][version] for d in dates],
                mode="lines+markers",
                name=f"v{version}",
            ),
            row=1,
            col=1,
        )

    fig.add_trace(
        go.Scatter(
            x=list(accumulated.keys()),
            y=list(accumulated.values()),
            mode="lines+markers",
            name="Accumulated",
        ),
        row=1,
        col=2,
    )

    def sort_key(v: str) -> tuple[int, ...]:
        return tuple(map(int, v.split(".")))

    xy = [(x, y) for x, y in sorted(totals.items(), key=lambda i: sort_key(i[0]))]

    fig.add_trace(
        go.Bar(
            x=[f"v{x}" for x, _ in xy],
            y=[y for _, y in xy],
            name="Total Downloads",
        ),
        row=2,
        col=1,
    )

    fig.update_xaxes(title_text="Date", row=1, col=1)
    fig.update_xaxes(title_text="Date", row=1, col=2)
    fig.update_xaxes(title_text="Version", row=2, col=1)

    fig.update_yaxes(title_text="Count", row=1, col=1)
    fig.update_yaxes(title_text="Accumulated", row=1, col=2)
    fig.update_yaxes(title_text="Downloads", row=2, col=1)

    fig.update_layout(
        title={"text": f"Downloads Analysis: {data.id}", "x": 0.5},
        template="plotly_white",
        margin={"l": 40, "r": 40, "t": 80, "b": 40},
    )
    fig.show()


if __name__ == "__main__":
    plot_downloads(fetch())



from __future__ import annotations

import uuid
from datetime import datetime, timezone
from itertools import count

from pgqueuer.models import Job


def mocked_job(
    id: int | count = count(),
    priority: int = 1,
    created: datetime | None = None,
    heartbeat: datetime | None = None,
    execute_after: datetime | None = None,
    updated: datetime | None = None,
    status: str = "queued",
    entrypoint: str = "test",
    payload: bytes | None = None,
    queue_manager_id: None | uuid.UUID = None,
) -> Job:
    now = datetime.now(timezone.utc)
    return Job(
        id=id if isinstance(id, int) else next(id),
        priority=priority,
        created=created or now,
        heartbeat=heartbeat or now,
        execute_after=execute_after or now,
        updated=updated or now,
        status=status,
        entrypoint=entrypoint,
        payload=payload,
        queue_manager_id=queue_manager_id or uuid.uuid4(),
    )



from __future__ import annotations

from datetime import timedelta

from pgqueuer.query_helpers import NormedEnqueueParam, normalize_enqueue_params


def test_normalize_single_entrypoint() -> None:
    result = normalize_enqueue_params(
        entrypoint="task1",
        payload=b"data1",
        priority=1,
        execute_after=timedelta(seconds=10),
        dedupe_key="key",
    )
    expected = NormedEnqueueParam(
        priority=[1],
        entrypoint=["task1"],
        payload=[b"data1"],
        execute_after=[timedelta(seconds=10)],
        dedupe_key=["key"],
    )
    assert result == expected


def test_normalize_multiple_entrypoints() -> None:
    result = normalize_enqueue_params(
        entrypoint=["task1", "task2"],
        payload=[b"data1", b"data2"],
        priority=[1, 2],
        execute_after=[timedelta(seconds=10), timedelta(seconds=20)],
        dedupe_key=["k1", "k2"],
    )
    expected = NormedEnqueueParam(
        priority=[1, 2],
        entrypoint=["task1", "task2"],
        payload=[b"data1", b"data2"],
        execute_after=[timedelta(seconds=10), timedelta(seconds=20)],
        dedupe_key=["k1", "k2"],
    )
    assert result == expected


def test_normalize_single_entrypoint_no_execute_after() -> None:
    result = normalize_enqueue_params(
        entrypoint="task1",
        payload=b"data1",
        priority=1,
    )
    expected = NormedEnqueueParam(
        priority=[1],
        entrypoint=["task1"],
        payload=[b"data1"],
        execute_after=[timedelta(seconds=0)],
        dedupe_key=[None],
    )
    assert result == expected


def test_normalize_single_entrypoint_no_dedupe_key() -> None:
    result = normalize_enqueue_params(
        entrypoint="task1",
        payload=b"data1",
        priority=1,
    )
    expected = NormedEnqueueParam(
        priority=[1],
        entrypoint=["task1"],
        payload=[b"data1"],
        execute_after=[timedelta(seconds=0)],
        dedupe_key=[None],
    )
    assert result == expected


def test_normalize_multiple_entrypoints_no_execute_after() -> None:
    result = normalize_enqueue_params(
        entrypoint=["task1", "task2"],
        payload=[b"data1", b"data2"],
        priority=[1, 2],
    )
    expected = NormedEnqueueParam(
        priority=[1, 2],
        entrypoint=["task1", "task2"],
        payload=[b"data1", b"data2"],
        execute_after=[timedelta(seconds=0), timedelta(seconds=0)],
        dedupe_key=[None, None],
    )
    assert result == expected


def test_normalize_mixed_execute_after() -> None:
    result = normalize_enqueue_params(
        entrypoint=["task1", "task2"],
        payload=[b"data1", b"data2"],
        priority=[1, 2],
        execute_after=[timedelta(seconds=10), None],
    )
    expected = NormedEnqueueParam(
        priority=[1, 2],
        entrypoint=["task1", "task2"],
        payload=[b"data1", b"data2"],
        execute_after=[timedelta(seconds=10), timedelta(seconds=0)],
        dedupe_key=[None, None],
    )
    assert result == expected


def test_normalize_mixed_dedupe_key() -> None:
    result = normalize_enqueue_params(
        entrypoint=["task1", "task2"],
        payload=[b"data1", b"data2"],
        priority=[1, 2],
        dedupe_key=["foo", "bar"],
    )
    expected = NormedEnqueueParam(
        priority=[1, 2],
        entrypoint=["task1", "task2"],
        payload=[b"data1", b"data2"],
        execute_after=[timedelta(seconds=0), timedelta(seconds=0)],
        dedupe_key=["foo", "bar"],
    )
    assert result == expected


def test_normalize_single_entrypoint_none_payload() -> None:
    result = normalize_enqueue_params(
        entrypoint="task1",
        payload=None,
        priority=1,
        execute_after=timedelta(seconds=10),
        dedupe_key="foo",
    )
    expected = NormedEnqueueParam(
        priority=[1],
        entrypoint=["task1"],
        payload=[None],
        execute_after=[timedelta(seconds=10)],
        dedupe_key=["foo"],
    )
    assert result == expected


def test_normalize_multiple_entrypoints_none_payload() -> None:
    result = normalize_enqueue_params(
        entrypoint=["task1", "task2"],
        payload=[None, b"data2"],
        priority=[1, 2],
        execute_after=[timedelta(seconds=10), timedelta(seconds=20)],
        dedupe_key=["foo", "bar"],
    )
    expected = NormedEnqueueParam(
        priority=[1, 2],
        entrypoint=["task1", "task2"],
        payload=[None, b"data2"],
        execute_after=[timedelta(seconds=10), timedelta(seconds=20)],
        dedupe_key=["foo", "bar"],
    )
    assert result == expected



from __future__ import annotations

import sys
from typing import Any, Callable
from unittest.mock import Mock

import pytest

from pgqueuer.factories import load_factory, run_factory


def test_load_factory_with_module_colon_function(monkeypatch: pytest.MonkeyPatch) -> None:
    """
    Test loading a factory using the 'module:function' syntax.
    """
    factory_path: str = "test_module:test_factory"

    # Create a mock factory function
    mock_factory: Mock = Mock(return_value="factory_result")
    # Create a mock module and assign the mock factory to it
    mock_module: Mock = Mock(spec_set=["test_factory"])
    mock_module.test_factory = mock_factory

    # Define a mock import_module function
    def mock_import_module(module_name: str) -> Any:
        if module_name == "test_module":
            return mock_module
        raise ImportError(f"No module named '{module_name}'")

    # Patch 'importlib.import_module' with the mock_import_module
    monkeypatch.setattr("importlib.import_module", mock_import_module)

    # Call the function under test
    factory: Callable[..., Any] = load_factory(factory_path)

    # Assertions
    assert factory is mock_factory, "The loaded factory should match the mock factory."


def test_load_factory_nonexistent_module(monkeypatch: pytest.MonkeyPatch) -> None:
    """
    Test loading a factory from a non-existent module.
    Expect an ImportError.
    """
    factory_path: str = "nonexistent_module:test_factory"

    # Define a mock import_module function that raises ImportError
    def mock_import_module(module_name: str) -> Any:
        raise ImportError(f"No module named '{module_name}'")

    # Patch 'importlib.import_module' with the mock_import_module
    monkeypatch.setattr("importlib.import_module", mock_import_module)

    with pytest.raises(ImportError, match="No module named 'nonexistent_module'"):
        load_factory(factory_path)


def test_load_factory_nonexistent_function(monkeypatch: pytest.MonkeyPatch) -> None:
    """
    Test loading a non-existent factory function from an existing module.
    Expect an AttributeError.
    """
    factory_path: str = "test_module:nonexistent_factory"

    # Create a mock module with only 'test_factory' attribute
    mock_module: Mock = Mock(spec_set=["test_factory"])
    mock_module.test_factory = Mock(return_value="factory_result")

    # Define a mock import_module function
    def mock_import_module(module_name: str) -> Any:
        if module_name == "test_module":
            return mock_module
        raise ImportError(f"No module named '{module_name}'")

    # Patch 'importlib.import_module' with the mock_import_module
    monkeypatch.setattr("importlib.import_module", mock_import_module)

    with pytest.raises(AttributeError, match="Mock object has no attribute 'nonexistent_factory'"):
        load_factory(factory_path)


def test_load_factory_sys_path_modified(monkeypatch: pytest.MonkeyPatch) -> None:
    """
    Ensure that the current working directory is inserted into sys.path.
    """
    factory_path: str = "test_module:test_factory"

    # Create a mock factory function
    mock_factory: Mock = Mock(return_value="factory_result")
    # Create a mock module and assign the mock factory to it
    mock_module: Mock = Mock(spec_set=["test_factory"])
    mock_module.test_factory = mock_factory

    # Define a mock import_module function
    def mock_import_module(module_name: str) -> Any:
        if module_name == "test_module":
            return mock_module
        raise ImportError(f"No module named '{module_name}'")

    # Patch 'importlib.import_module' and 'os.getcwd'
    monkeypatch.setattr("importlib.import_module", mock_import_module)
    monkeypatch.setattr("os.getcwd", lambda: "/fake/dir")

    # Capture the original sys.path
    original_sys_path: list[str] = list(sys.path)

    try:
        # Call the function under test
        load_factory(factory_path)

        # Assertions
        assert sys.path[0] == "/fake/dir", (
            "The current working directory should be inserted at sys.path[0]."
        )
    finally:
        # Restore sys.path to its original state
        sys.path = original_sys_path


async def test_run_factory_with_async_context_manager() -> None:
    """
    Test run_factory with an asynchronous context manager.
    """
    from contextlib import AbstractAsyncContextManager

    class AsyncCM(Mock, AbstractAsyncContextManager[Any]):
        async def __aenter__(self) -> str:
            return "async_cm_value"

        async def __aexit__(self, exc_type: Any, exc: Any, tb: Any) -> None:
            pass

    async_cm: AsyncCM = AsyncCM()

    async with run_factory(async_cm) as value:
        assert value == "async_cm_value", (
            "The async context manager should yield the correct value."
        )


async def test_run_factory_with_sync_context_manager(monkeypatch: pytest.MonkeyPatch) -> None:
    """
    Test run_factory with a synchronous context manager.
    """
    from contextlib import AbstractContextManager

    class SyncCM(Mock, AbstractContextManager[Any]):
        def __enter__(self) -> str:
            return "sync_cm_value"

        def __exit__(self, exc_type: Any, exc: Any, tb: Any) -> None:
            pass

    sync_cm: SyncCM = SyncCM()

    async with run_factory(sync_cm) as value:
        assert value == "sync_cm_value", "The sync context manager should yield the correct value."


async def test_run_factory_with_awaitable() -> None:
    """
    Test run_factory with an awaitable.
    """

    async def sample_awaitable() -> str:
        return "awaitable_value"

    awaitable: Any = sample_awaitable()

    async with run_factory(awaitable) as value:
        assert value == "awaitable_value", "The awaitable should yield the awaited value."


async def test_run_factory_with_invalid_input() -> None:
    """
    Test run_factory with an invalid input that is neither an awaitable nor a context manager.
    Expect a TypeError.
    """
    invalid_input: str = "not a valid input"

    with pytest.raises(TypeError, match="object str can't be used in 'await' expression"):
        async with run_factory(invalid_input):  # type: ignore[arg-type]
            pass  # This line should not be reached



from __future__ import annotations

from contextlib import asynccontextmanager
from datetime import datetime
from typing import AsyncGenerator

import asyncpg

from pgqueuer import PgQueuer
from pgqueuer.db import AsyncpgDriver
from pgqueuer.logconfig import logger
from pgqueuer.models import Job, Schedule


async def create_pgqueuer() -> PgQueuer:
    connection = await asyncpg.connect()
    driver = AsyncpgDriver(connection)
    pgq = PgQueuer(driver)

    # Setup the 'fetch' entrypoint
    @pgq.entrypoint("fetch")
    async def process_message(job: Job) -> None:
        print(f"Processed message: {job!r}")

    @pgq.schedule("scheduled_every_minute", "* * * * *")
    async def scheduled_every_minute(schedule: Schedule) -> None:
        print(f"Executed every minute {schedule!r} {datetime.now()!r}")

    return pgq


@asynccontextmanager
async def main() -> AsyncGenerator[PgQueuer, None]:
    """
    A context manager for setting up and tearing down the PgQueuer instance.

    This function manages the lifecycle of the PgQueuer instance, ensuring proper setup and teardown
    when used in an asynchronous context. It includes the following steps:

    Setup:
        - Logs the start of the setup process.
        - Creates and configures a PgQueuer instance by connecting to the database and initializing
          entrypoints and schedules.

    Teardown:
        - Logs the start of the teardown process.
        - Ensures cleanup actions for the PgQueuer instance, releasing resources like database
          connections.

    Yields:
        PgQueuer: The configured instance ready for processing jobs and schedules.
    """
    logger.info("setup")
    try:
        pgq = await create_pgqueuer()
        yield pgq
    finally:
        logger.info("teardown")



"""
Logging configuration for the pgqueuer application.
"""

from __future__ import annotations

import logging
import logging.config
import sys
from datetime import datetime
from enum import Enum
from typing import Final


class LogLevel(Enum):
    CRITICAL = "CRITICAL"
    FATAL = "FATAL"
    ERROR = "ERROR"
    WARN = "WARN"
    WARNING = "WARNING"
    INFO = "INFO"
    DEBUG = "DEBUG"
    NOTSET = "NOTSET"


class ISOFormatter(logging.Formatter):
    """Outputs log timestamps in ISO 8601 format with timezone information."""

    def formatTime(self, record: logging.LogRecord, datefmt: str | None = None) -> str:
        return datetime.fromtimestamp(record.created).astimezone().isoformat()


class MaxLevelFilter(logging.Filter):
    """Only allows log records up to a specified maximum logging level."""

    def __init__(self, max_level: int):
        super().__init__()
        self.max_level = max_level

    def filter(self, record: logging.LogRecord) -> bool:
        return record.levelno <= self.max_level


def setup_fancy_logger(level: LogLevel) -> None:
    """
    Update the 'pgqueuer' logger's level dynamically.
    """

    logging.config.dictConfig(
        {
            "version": 1,
            "disable_existing_loggers": False,
            "formatters": {
                "default": {
                    "()": ISOFormatter,
                    "format": "%(asctime)s [%(name)s] %(levelname)s: %(message)s",
                },
            },
            "filters": {
                "max_info": {
                    "()": MaxLevelFilter,
                    "max_level": logging.INFO,
                },
            },
            "handlers": {
                "stdout": {
                    "class": "logging.StreamHandler",
                    "level": "DEBUG",
                    "formatter": "default",
                    "stream": sys.stdout,
                    "filters": ["max_info"],
                },
                "stderr": {
                    "class": "logging.StreamHandler",
                    "level": "WARNING",
                    "formatter": "default",
                    "stream": sys.stderr,
                },
            },
            "loggers": {
                "pgqueuer": {
                    "handlers": ["stdout", "stderr"],
                    "level": level.name,
                }
            },
        }
    )


# Create the 'pgqueuer' logger and attach a NullHandler by default.
logger: Final = logging.getLogger("pgqueuer")
logger.addHandler(logging.NullHandler())



import asyncio
import functools
from datetime import timedelta
from multiprocessing import Process, Queue as MPQueue

import anyio
import pytest
from helpers import mocked_job

from pgqueuer.db import Driver
from pgqueuer.errors import MaxRetriesExceeded, MaxTimeExceeded
from pgqueuer.executors import (
    AbstractEntrypointExecutor,
    EntrypointExecutor,
    EntrypointExecutorParameters,
    RetryWithBackoffEntrypointExecutor,
    is_async_callable,
)
from pgqueuer.helpers import timer
from pgqueuer.models import Channel, Context, Job
from pgqueuer.qm import QueueManager
from pgqueuer.queries import Queries


class MultiprocessingExecutor(AbstractEntrypointExecutor):
    def __init__(self) -> None:
        self.requests_per_second = 5
        self.retry_timer = timedelta(seconds=10)
        self.serialized_dispatch = False
        self.concurrency_limit = 2
        self.queue: MPQueue[object] = MPQueue()

    async def execute(self, job: Job, context: Context) -> None:
        process = Process(target=self.process_function, args=(job,))
        process.start()
        process.join()

    def process_function(self, job: Job) -> None:
        # Simulate processing and put result in queue
        if job.payload:
            self.queue.put(job.payload)


@pytest.mark.asyncio
async def test_entrypoint_executor_sync(apgdriver: Driver) -> None:
    result = []

    def sync_function(job: Job) -> None:
        if job.payload:
            result.append(job.payload)

    executor = EntrypointExecutor(
        EntrypointExecutorParameters(
            channel=Channel("foo"),
            concurrency_limit=10,
            connection=apgdriver,
            queries=Queries(apgdriver),
            requests_per_second=float("+inf"),
            retry_timer=timedelta(seconds=300),
            serialized_dispatch=False,
            shutdown=asyncio.Event(),
            func=sync_function,
        )
    )
    job = mocked_job(payload=b"test_payload")

    await executor.execute(
        job,
        Context(anyio.CancelScope()),
    )

    assert result == [b"test_payload"]


@pytest.mark.asyncio
async def test_entrypoint_executor_async(apgdriver: Driver) -> None:
    result = []

    async def async_function(job: Job) -> None:
        if job.payload:
            result.append(job.payload)

    executor = EntrypointExecutor(
        EntrypointExecutorParameters(
            channel=Channel("foo"),
            concurrency_limit=10,
            connection=apgdriver,
            queries=Queries(apgdriver),
            requests_per_second=float("+inf"),
            retry_timer=timedelta(seconds=300),
            serialized_dispatch=False,
            shutdown=asyncio.Event(),
            func=async_function,
        )
    )
    job = mocked_job(payload=b"test_payload")

    await executor.execute(
        job,
        Context(anyio.CancelScope()),
    )
    assert result == [b"test_payload"]


@pytest.mark.asyncio
async def test_custom_threading_executor() -> None:
    class ThreadingExecutor(AbstractEntrypointExecutor):
        def __init__(self) -> None:
            self.requests_per_second = 10
            self.retry_timer = timedelta(seconds=5)
            self.serialized_dispatch = False
            self.concurrency_limit = 5
            self.result = list[bytes]()

        async def execute(self, job: Job, context: Context) -> None:
            loop = asyncio.get_event_loop()
            await loop.run_in_executor(None, self.threaded_function, job)

        def threaded_function(self, job: Job) -> None:
            if job.payload:
                self.result.append(job.payload)

    executor = ThreadingExecutor()
    job = mocked_job(payload=b"thread_payload")

    await executor.execute(
        job,
        Context(anyio.CancelScope()),
    )

    assert executor.result == [b"thread_payload"]


@pytest.mark.asyncio
async def test_custom_multiprocessing_executor() -> None:
    executor = MultiprocessingExecutor()
    job = mocked_job(payload=b"process_payload")

    await executor.execute(
        job,
        Context(anyio.CancelScope()),
    )
    result = executor.queue.get()

    assert result == b"process_payload"


@pytest.mark.asyncio
async def test_queue_manager_with_custom_executor(apgdriver: Driver) -> None:
    qm = QueueManager(connection=apgdriver)
    results = []

    class CustomExecutor(AbstractEntrypointExecutor):
        async def execute(self, job: Job, context: Context) -> None:
            if job.payload:
                results.append(job.payload)

    @qm.entrypoint(
        name="custom_entrypoint",
        executor_factory=CustomExecutor,
    )
    def entrypoint_function(job: Job) -> None:
        pass  # Not used since executor handles execution

    queries = Queries(apgdriver)
    await queries.enqueue(entrypoint="custom_entrypoint", payload=b"test_data")

    async def run_queue_manager() -> None:
        await qm.run(dequeue_timeout=timedelta(seconds=2), batch_size=1)

    task = asyncio.create_task(run_queue_manager())
    await asyncio.sleep(1)  # Allow some time for the job to be processed
    qm.shutdown.set()  # Signal the queue manager to stop
    await task

    assert results == [b"test_data"]


async def async_function(job: Job) -> None:
    await asyncio.sleep(0)


def sync_function(job: Job) -> None:
    pass


def test_is_async_callable_with_async_function() -> None:
    assert is_async_callable(async_function) is True


def test_is_async_callable_with_sync_function() -> None:
    assert is_async_callable(sync_function) is False


def test_is_async_callable_with_partial_async_function() -> None:
    partial_async = functools.partial(async_function)
    assert is_async_callable(partial_async) is True


def test_is_async_callable_with_partial_sync_function() -> None:
    partial_sync = functools.partial(sync_function)
    assert is_async_callable(partial_sync) is False


def test_is_async_callable_with_class_method() -> None:
    class MyClass:
        async def async_method(self, job: Job) -> None:
            await asyncio.sleep(0)

        def sync_method(self, job: Job) -> None:
            pass

    instance = MyClass()
    assert is_async_callable(instance.async_method) is True
    assert is_async_callable(instance.sync_method) is False


async def test_retry_with_backoff_entrypoint_executor_max_attempts(apgdriver: Driver) -> None:
    jobs = list[Job]()

    async def raises(job: Job) -> None:
        await asyncio.sleep(0)
        jobs.append(job)
        raise ValueError

    parameters = EntrypointExecutorParameters(
        channel=Channel("test_retry_with_backoff_entrypoint_executor_max_attempts"),
        concurrency_limit=10,
        connection=apgdriver,
        queries=Queries(apgdriver),
        requests_per_second=float("+inf"),
        retry_timer=timedelta(seconds=300),
        serialized_dispatch=False,
        shutdown=asyncio.Event(),
        func=raises,
    )
    exc = RetryWithBackoffEntrypointExecutor(
        parameters,
        initial_delay=0,
        jitter=lambda: 0,
    )

    exc.max_attempts = 5
    exc.max_time = timedelta(seconds=300)
    mj = mocked_job()
    with pytest.raises(MaxRetriesExceeded):
        await exc.execute(mj, Context(anyio.CancelScope()))
    assert sum(j.id == mj.id for j in jobs) == exc.max_attempts

    exc.max_attempts = 10
    exc.max_time = timedelta(seconds=300)
    mj = mocked_job()
    with pytest.raises(MaxRetriesExceeded):
        await exc.execute(mj, Context(anyio.CancelScope()))
    assert sum(j.id == mj.id for j in jobs) == exc.max_attempts


async def test_retry_with_backoff_entrypoint_executor_max_time(apgdriver: Driver) -> None:
    async def raises(_: Job) -> None:
        await asyncio.sleep(0.01)
        raise ValueError

    parameters = EntrypointExecutorParameters(
        channel=Channel("test_retry_with_backoff_entrypoint_executor_max_time"),
        concurrency_limit=10,
        connection=apgdriver,
        queries=Queries(apgdriver),
        requests_per_second=float("+inf"),
        retry_timer=timedelta(seconds=300),
        serialized_dispatch=False,
        shutdown=asyncio.Event(),
        func=raises,
    )
    exc = RetryWithBackoffEntrypointExecutor(
        parameters,
        initial_delay=0,
        jitter=lambda: 0,
    )

    exc.max_attempts = 1000
    exc.max_time = timedelta(seconds=0.01)
    mj = mocked_job()
    with (
        timer() as elp,
        pytest.raises(MaxTimeExceeded),
    ):
        await exc.execute(mj, Context(anyio.CancelScope()))

    leeway = 1.1
    assert elp() * leeway >= exc.max_time

    exc.max_attempts = 1000
    exc.max_time = timedelta(seconds=0.1)
    mj = mocked_job()
    with (
        timer() as elp,
        pytest.raises(MaxTimeExceeded),
    ):
        await exc.execute(mj, Context(anyio.CancelScope()))

    assert elp() * leeway >= exc.max_time


async def test_retry_with_backoff_entrypoint_executor_until_pass(apgdriver: Driver) -> None:
    N = 10
    jobs = list[Job]()

    async def raises(job: Job) -> None:
        await asyncio.sleep(0.001)
        jobs.append(job)
        if len(jobs) > N:
            return
        raise ValueError

    parameters = EntrypointExecutorParameters(
        channel=Channel("test_retry_with_backoff_entrypoint_executor_until_pass"),
        concurrency_limit=10,
        connection=apgdriver,
        queries=Queries(apgdriver),
        requests_per_second=float("+inf"),
        retry_timer=timedelta(seconds=300),
        serialized_dispatch=False,
        shutdown=asyncio.Event(),
        func=raises,
    )
    exc = RetryWithBackoffEntrypointExecutor(
        parameters,
        initial_delay=0,
        jitter=lambda: 0,
    )

    exc.max_attempts = 1000
    exc.max_time = timedelta(seconds=0.1)
    mj = mocked_job()
    await exc.execute(mj, Context(anyio.CancelScope()))



"""
Module for buffering items and their statuses before processing.

This module defines the `TimedOverflowBuffer` class, which accumulates items and their statuses
until either a specified capacity is reached or a timeout occurs. It then flushes the
buffer by invoking a provided asynchronous callable. This buffering mechanism helps reduce the
number of processing operations by batching them, improving performance.
"""

from __future__ import annotations

import asyncio
import dataclasses
import time
from contextlib import suppress
from datetime import timedelta
from typing import (
    AsyncGenerator,
    Awaitable,
    Callable,
    Generic,
    TypeVar,
)

from typing_extensions import Self

from . import helpers, logconfig, models, tm

T = TypeVar("T")


@dataclasses.dataclass
class TimedOverflowBuffer(Generic[T]):
    """
    Accumulates items, flushing them based on timeouts or buffer capacity.

    The `TimedOverflowBuffer` class collects items in a buffer and flushes them
    when either the maximum number of items (`max_size`) is reached or a specified
    timeout (`timeout`) has elapsed since the last flush. Flushing involves invoking
    an asynchronous callback with the accumulated items.

    The class includes mechanisms for retrying failed flush operations and ensures
    a graceful shutdown by attempting to flush remaining items before exit.

    Attributes:
        max_size (int): The maximum number of items to buffer before triggering a flush.
        timeout (timedelta): The maximum duration to wait before automatically flushing
            the buffer, regardless of size.
        callback (Callable[[list[T]], Awaitable[None]]): The asynchronous callable invoked
            during a flush operation to process the buffered items.
        retry_backoff (helpers.ExponentialBackoff): A backoff strategy used to retry
            failed flush operations during normal operation.
        shutdown_backoff (helpers.ExponentialBackoff): A backoff strategy used during
            shutdown to ensure all items are flushed before the process exits.
        shutdown (asyncio.Event): An event that signals when the buffer should stop
            operations, such as during shutdown.
        events (asyncio.Queue[T]): An asynchronous queue holding the buffered items.
        lock (asyncio.Lock): A lock to prevent concurrent flush operations.
        tm (tm.TaskManager): A task manager for managing background tasks associated
            with the buffer's operations.
    """

    max_size: int
    callback: Callable[[list[T]], Awaitable[None]]
    timeout: timedelta = dataclasses.field(
        default_factory=lambda: timedelta(seconds=0.1),
    )
    retry_backoff: helpers.ExponentialBackoff = dataclasses.field(
        default_factory=lambda: helpers.ExponentialBackoff(
            start_delay=timedelta(seconds=0.01),
            max_delay=timedelta(seconds=10),
        ),
    )
    shutdown_backoff: helpers.ExponentialBackoff = dataclasses.field(
        default_factory=lambda: helpers.ExponentialBackoff(
            start_delay=timedelta(milliseconds=1),
            max_delay=timedelta(milliseconds=100),
        )
    )

    shutdown: asyncio.Event = dataclasses.field(
        init=False,
        default_factory=asyncio.Event,
    )
    events: asyncio.Queue[T] = dataclasses.field(
        init=False,
        default_factory=asyncio.Queue,
    )
    lock: asyncio.Lock = dataclasses.field(
        init=False,
        default_factory=asyncio.Lock,
    )
    tm: tm.TaskManager = dataclasses.field(
        init=False,
        default_factory=tm.TaskManager,
    )

    async def periodic_flush(self) -> None:
        while not self.shutdown.is_set():
            if not self.lock.locked() and self.events.qsize() > 0:
                self.tm.add(asyncio.create_task(self.flush()))

            # await asyncio.sleep(helpers.timeout_with_jitter(self.timeout).total_seconds())
            with suppress(asyncio.TimeoutError, TimeoutError):
                await asyncio.wait_for(
                    asyncio.create_task(self.shutdown.wait()),
                    timeout=helpers.timeout_with_jitter(self.timeout).total_seconds(),
                )

    async def add(self, item: T) -> None:
        """
        Add an item to the buffer; flush if buffer reaches maximum size.

        This method adds an item to the internal events queue.
        If the number of events in the buffer reaches or exceeds `max_size`, it
        triggers a flush of the buffer to invoke the provided asynchronous flush callable.
        Additionally, it schedules a flush operation to occur after the specified timeout.

        Args:
            item (T): The item to be added to the buffer.
        """
        await self.events.put(item)

        if self.events.qsize() >= self.max_size and not self.lock.locked():
            self.tm.add(asyncio.create_task(self.flush()))

    async def pop_until(
        self,
        until: timedelta = timedelta(seconds=0.01),
    ) -> AsyncGenerator[T, None]:
        """
        Yield items from the buffer until conditions are met.

        This asynchronous generator yields items from the internal events queue.
        It continues to yield items until either the queue is empty or the elapsed time since
        starting exceeds twice the buffer's timeout. This helps prevent the flush operation
        from taking too long or processing too many items at once.

        Yields:
            AsyncGenerator[T, None]: An asynchronous generator yielding items.
        """
        deadline = time.time() + until.total_seconds()
        while not self.events.empty() and time.time() < deadline:
            yield await self.events.get()

    async def flush(self) -> None:
        """
        Flush the accumulated items in the buffer by invoking the provided asynchronous callable.

        Collects all items currently in the buffer by consuming the events from
        the internal queue using `pop_until`. If there are any events, it attempts to invoke the
        provided asynchronous callable with the events. If an exception occurs during the invocation
        of the callable, it logs the exception, re-adds the items to the queue, and schedules
        a retry. This helps in handling transient errors without losing items.
        """

        if self.lock.locked():
            return

        async with self.lock:
            items = [item async for item in self.pop_until()]

            if not items:
                return

            try:
                await self.callback(items)
            except Exception as e:
                delay = self.retry_backoff.next_delay()
                logconfig.logger.warning(
                    "Unable to flush(%s): %s\nRetry in: %r",
                    self.callback.__name__,
                    str(e),
                    delay,
                )
                # Re-add the items to the queue for retry
                for item in items:
                    self.events.put_nowait(item)

                await asyncio.sleep(
                    0
                    if self.shutdown.is_set()
                    else helpers.timeout_with_jitter(delay).total_seconds()
                )

            else:
                self.retry_backoff.reset()

    async def __aenter__(self) -> Self:
        """
        Enter the asynchronous context manager.

        Returns the TimedOverflowBuffer instance itself, allowing it to
        be used within an `async with` block.

        Returns:
            TimedOverflowBuffer: The TimedOverflowBuffer instance itself.
        """
        self.tm.add(asyncio.create_task(self.periodic_flush()))
        return self

    async def __aexit__(self, *_: object) -> None:
        """
        Exit the asynchronous context manager, ensuring all items are flushed.

        This method is called when exiting an `async with` block. It ensures the buffer
        performs any final flush operations for the accumulated items before shutdown.
        The following steps are executed:

        1. Signals the buffer to stop periodic operations by setting the `shutdown` event.
        2. Waits for all ongoing tasks managed by the buffer to complete.
        3. Attempts to flush any remaining items in the buffer using the `shutdown_backoff`
           strategy to handle transient failures, such as temporary database outages.
        4. Stops retrying after the maximum backoff limit (`shutdown_backoff.max_limit`)
           is reached. This ensures that the application does not hang indefinitely
           during situations like prolonged database downtime or critical errors.

        Note:
            This mechanism ensures that the application attempts to process all items
            gracefully before shutting down. However, if the flush operation repeatedly
            fails (e.g., due to a database outage), the method will eventually give up
            to prevent blocking the application indefinitely.
        """

        self.shutdown.set()
        await self.tm.gather_tasks()

        while self.shutdown_backoff.current_delay < self.shutdown_backoff.max_delay:
            await self.flush()
            if self.events.empty():
                break
            await asyncio.sleep(self.shutdown_backoff.next_delay().total_seconds())


class JobStatusLogBuffer(
    TimedOverflowBuffer[
        tuple[
            models.Job,
            models.JOB_STATUS,
            models.TracebackRecord | None,
        ]
    ]
):
    """
    Specialized TimedOverflowBuffer for handling Job/Status-log.
    """


class HeartbeatBuffer(TimedOverflowBuffer[models.JobId]):
    """
    Specialized TimedOverflowBuffer for handling heartbeats.
    """


class RequestsPerSecondBuffer(TimedOverflowBuffer[str]):
    """
    Specialized TimedOverflowBuffer for handling RPS.
    """



import pytest

from pgqueuer.qb import Durability, DurabilityPolicy


@pytest.mark.parametrize(
    "durability,expected_policy",
    [
        (
            Durability.volatile,
            DurabilityPolicy(
                queue_table="UNLOGGED",
                queue_log_table="UNLOGGED",
                statistics_table="UNLOGGED",
                schedules_table="UNLOGGED",
            ),
        ),
        (
            Durability.balanced,
            DurabilityPolicy(
                queue_table="",
                queue_log_table="UNLOGGED",
                statistics_table="UNLOGGED",
                schedules_table="",
            ),
        ),
        (
            Durability.durable,
            DurabilityPolicy(
                queue_table="",
                queue_log_table="",
                statistics_table="",
                schedules_table="",
            ),
        ),
    ],
)
def test_durability_policy(durability: Durability, expected_policy: DurabilityPolicy) -> None:
    """
    Test that the `config` property of each `Durability` level correctly returns
    the expected `DurabilityPolicy`.
    """
    assert durability.config == expected_policy


@pytest.mark.parametrize(
    "policy,expected_logged",
    [
        (
            DurabilityPolicy(
                queue_table="",
                queue_log_table="UNLOGGED",
                statistics_table="UNLOGGED",
                schedules_table="",
            ),
            {
                "pgqueuer": True,
                "pgqueuer_log": False,
                "pgqueuer_statistics": False,
                "pgqueuer_schedules": True,
            },
        ),
        (
            DurabilityPolicy(
                queue_table="UNLOGGED",
                queue_log_table="UNLOGGED",
                statistics_table="UNLOGGED",
                schedules_table="UNLOGGED",
            ),
            {
                "pgqueuer": False,
                "pgqueuer_log": False,
                "pgqueuer_statistics": False,
                "pgqueuer_schedules": False,
            },
        ),
    ],
)
def test_policy_attributes(policy: DurabilityPolicy, expected_logged: dict[str, bool]) -> None:
    """
    Test that individual attributes in `DurabilityPolicy` correctly indicate whether they
    are logged or unlogged.
    """
    assert (policy.queue_table == "") == expected_logged["pgqueuer"]
    assert (policy.queue_log_table == "") == expected_logged["pgqueuer_log"]
    assert (policy.statistics_table == "") == expected_logged["pgqueuer_statistics"]
    assert (policy.schedules_table == "") == expected_logged["pgqueuer_schedules"]


def test_invalid_durability_level() -> None:
    """
    Test that accessing an unknown durability level raises a ValueError.
    """
    with pytest.raises(ValueError, match="'nonexistent' is not a valid Durability"):
        Durability("nonexistent")



import asyncio
import uuid
from datetime import timedelta
from itertools import count

import pytest
from helpers import mocked_job

from pgqueuer.buffers import JobStatusLogBuffer
from pgqueuer.models import JOB_STATUS, Job


def job_faker(
    cnt: count = count(),
    queue_manager_id: uuid.UUID = uuid.uuid4(),
) -> Job:
    return mocked_job(
        id=next(cnt),
        status="picked",
        entrypoint="foo",
        priority=0,
        payload=None,
        queue_manager_id=queue_manager_id,
    )


@pytest.mark.parametrize("max_size", (1, 2, 3, 5, 64))
async def test_job_buffer_max_size(max_size: int) -> None:
    helper_buffer = []

    async def helper(x: list) -> None:
        helper_buffer.extend(x)

    async with JobStatusLogBuffer(
        max_size=max_size,
        timeout=timedelta(seconds=100),
        callback=helper,
    ) as buffer:
        for _ in range(max_size - 1):
            await buffer.add((job_faker(), "successful", None))
            assert len(helper_buffer) == 0

        await buffer.add((job_faker(), "successful", None))

    # On ctx-mangner exit flush is forced.
    assert len(helper_buffer) == max_size


@pytest.mark.parametrize("N", (5, 64))
@pytest.mark.parametrize("timeout", (timedelta(seconds=0.01), timedelta(seconds=0.001)))
async def test_job_buffer_timeout(
    N: int,
    timeout: timedelta,
) -> None:
    helper_buffer = []

    async def helper(x: list) -> None:
        helper_buffer.extend(x)

    async with JobStatusLogBuffer(
        max_size=N * 2,
        timeout=timeout,
        callback=helper,
    ) as buffer:
        for _ in range(N):
            await buffer.add((job_faker(), "successful", None))
            assert len(helper_buffer) == 0

        await asyncio.sleep(timeout.total_seconds() * 1.1)

    assert len(helper_buffer) == N


@pytest.mark.parametrize("max_size", (2, 3, 5, 64))  # Adjusted to max_size >=2
async def test_job_buffer_flush_on_exit(max_size: int) -> None:
    """
    Test that the buffer flushes all remaining items upon exiting the context,
    even if max_size is not reached and timeout hasn't occurred.
    """
    helper_buffer = []

    async def helper(x: list) -> None:
        helper_buffer.extend(x)

    async with JobStatusLogBuffer(
        max_size=max_size,
        timeout=timedelta(seconds=100),
        callback=helper,
    ) as buffer:
        for _ in range(max_size - 2):
            await buffer.add((job_faker(), "successful", None))
            assert len(helper_buffer) == 0

    # After exiting the context, remaining items should be flushed
    assert len(helper_buffer) == max_size - 2


@pytest.mark.parametrize("max_size", (1, 2, 3, 5, 64))
@pytest.mark.parametrize("flushes", (1, 2, 3, 5, 64))
async def test_job_buffer_multiple_flushes(max_size: int, flushes: int) -> None:
    """
    Test that the buffer can handle multiple flushes when more items than max_size are added.
    """
    helper_buffer = []

    async def helper(x: list) -> None:
        helper_buffer.append(x)

    async with JobStatusLogBuffer(
        max_size=max_size,
        timeout=timedelta(seconds=100),
        callback=helper,
    ) as buffer:
        for _ in range(flushes):
            for _ in range(max_size):
                await buffer.add((job_faker(), "successful", None))
            await buffer.flush()

    # Verify that the buffer flushed three times
    assert len(helper_buffer) == flushes


@pytest.mark.parametrize("max_size", (1, 2, 3, 5, 64))
async def test_job_buffer_flush_on_exception(max_size: int) -> None:
    """
    Test that the buffer handles exceptions in the callback gracefully and retries.
    """
    helper_buffer = []
    flush_call_count = 0

    async def faulty_helper(x: list) -> None:
        nonlocal flush_call_count
        flush_call_count += 1
        if flush_call_count < 2:
            raise RuntimeError("Simulated flush failure")
        helper_buffer.extend(x)

    async with JobStatusLogBuffer(
        max_size=max_size,
        timeout=timedelta(seconds=0.01),
        callback=faulty_helper,
    ) as buffer:
        for _ in range(max_size):
            await buffer.add((job_faker(), "successful", None))

        # Allow time for the flush to be attempted and retried
        await asyncio.sleep(0.02)

    # The first flush should have failed, and the second should have succeeded
    assert flush_call_count == 2
    assert len(helper_buffer) == max_size


@pytest.mark.parametrize("max_size", (1, 2, 3, 5, 64))
async def test_job_buffer_flush_order(max_size: int) -> None:
    """
    Test that items are flushed in the order they were added.
    """
    helper_buffer = []

    async def helper(x: list) -> None:
        helper_buffer.extend(x)

    async with JobStatusLogBuffer(
        max_size=max_size,
        timeout=timedelta(seconds=100),
        callback=helper,
    ) as buffer:
        items = [(job_faker(), "successful") for _ in range(max_size)]
        for item in items:
            await buffer.add(item)  # type: ignore[arg-type]

    assert helper_buffer == items


@pytest.mark.parametrize("max_size", (1, 2, 3, 5, 64))
async def test_job_buffer_concurrent_adds(max_size: int) -> None:
    """
    Test that the buffer can handle concurrent additions without losing items.
    """
    helper_buffer = []

    async def helper(x: list) -> None:
        helper_buffer.extend(x)

    async with JobStatusLogBuffer(
        max_size=max_size,
        timeout=timedelta(seconds=100),
        callback=helper,
    ) as buffer:

        async def add_items(n: int) -> None:
            for _ in range(n):
                await buffer.add((job_faker(), "successful", None))

        tasks = [asyncio.create_task(add_items(max_size // 2)) for _ in range(4)]
        await asyncio.gather(*tasks)

    # Total items added: (max_size // 2) * 4
    expected = (max_size // 2) * 4
    assert len(helper_buffer) == expected


async def test_job_buffer_empty_flush() -> None:
    """
    Test that flushing an empty buffer does not cause any issues.
    """
    helper_buffer = []

    async def helper(x: list) -> None:
        helper_buffer.extend(x)

    async with JobStatusLogBuffer(
        max_size=10,
        timeout=timedelta(seconds=0.1),
        callback=helper,
    ):
        # Do not add any items and let the buffer flush on exit
        pass

    assert len(helper_buffer) == 0


@pytest.mark.parametrize("max_size", (1, 2, 3, 5, 64))
async def test_job_buffer_reuse_after_flush(max_size: int) -> None:
    """
    Test that the buffer can be reused after a flush has occurred.
    """
    helper_buffer = []

    async def helper(x: list) -> None:
        helper_buffer.extend(x)

    async with JobStatusLogBuffer(
        max_size=max_size,
        timeout=timedelta(seconds=100),
        callback=helper,
    ) as buffer:
        # First flush
        for _ in range(max_size):
            await buffer.add((job_faker(), "successful", None))
        await buffer.flush()
        assert len(helper_buffer) == max_size

        # Reset helper_buffer for the second flush
        helper_buffer.clear()

        # Second flush
        for _ in range(max_size):
            await buffer.add((job_faker(), "successful", None))
        await buffer.flush()
        assert len(helper_buffer) == max_size


@pytest.mark.parametrize("max_size", (1, 2, 3, 5, 64))
async def test_job_buffer_exception_during_flush(max_size: int) -> None:
    """
    Test that the buffer handles exceptions during flush without losing items.
    """
    helper_buffer = []
    flush_call_count = 0

    async def faulty_helper(x: list) -> None:
        nonlocal flush_call_count
        flush_call_count += 1
        if flush_call_count == 1:
            raise RuntimeError("Simulated flush failure")
        helper_buffer.extend(x)

    async with JobStatusLogBuffer(
        max_size=max_size,
        timeout=timedelta(seconds=0.01),
        callback=faulty_helper,
    ) as buffer:
        for _ in range(max_size):
            await buffer.add((job_faker(), "successful", None))

        # Allow time for the flush to be attempted and retried
        await asyncio.sleep(0.02)

    # After first failure, flush should retry and succeed
    assert flush_call_count == 2
    assert len(helper_buffer) == max_size


@pytest.mark.parametrize("max_size", (1, 2, 3, 5, 64))
async def test_job_buffer_callback_called_correctly(max_size: int) -> None:
    """
    Test that the callback is called with the correct items.
    """
    items = [(job_faker(), "successful") for _ in range(max_size)]
    received_items = []

    async def helper(x: list) -> None:
        received_items.extend(x)

    async with JobStatusLogBuffer(
        max_size=max_size,
        timeout=timedelta(seconds=100),
        callback=helper,
    ) as buffer:
        for item in items:
            await buffer.add(item)  # type: ignore[arg-type]

    assert received_items == items


async def test_job_buffer_callback_exception_during_teardown() -> None:
    N = 10
    items: list[tuple[Job, JOB_STATUS, None]] = [
        (job_faker(), "successful", None) for _ in range(N)
    ]

    async def helper(_: object) -> None:
        raise ValueError

    async with JobStatusLogBuffer(
        max_size=N**2,  # max size must be gt. N.
        timeout=timedelta(seconds=60),  # must be gt. run time of 'for loop' in the with block.
        callback=helper,
    ) as buffer:
        for item in items:
            await buffer.add(item)

    # Was uanble to flush at exit, buffer should have all elements.
    assert buffer.events.qsize() == N



"""PostgreSQL NOTIFY event‑handling helpers
================================================

This module replaces the legacy *if/elif* `handle_event_type` chain with a
small, type‑safe dispatch system based on :class:`EventRouter`.  Public
components
-----------------

``EventRouter``
    Lightweight dispatcher that maps a single ``event.type`` string to a
    handler.  Handlers receive the concrete ``models.*Event`` subclass
    selected by *Pydantic*’s ``discriminator`` and **must** return
    ``None``.

``build_default_router``
    Convenience factory that wires handlers for the four canonical
    back‑end events (*table‑changed*, *requests‑per‑second*,
    *cancellation*, *health‑check*).  It injects application state via
    closures, avoiding a bulky context object.

``PGNoticeEventListener``
    A thin ``asyncio.Queue`` subclass used by the default
    *table‑changed* handler to deliver row‑level notifications to
    consumers.

``initialize_notice_event_listener``
    Glue that attaches any ``Callable[[models.AnyEvent], None]``
    (typically the router) to a live database connection through
    ``connection.add_listener``.

Typical usage::

    router = build_default_router(...state injections...)
    await initialize_notice_event_listener(conn, "mychannel", router)
"""

from __future__ import annotations

import asyncio
import uuid
from collections.abc import Callable
from dataclasses import dataclass, field
from typing import MutableMapping, TypeAlias, TypeVar

from . import db, logconfig, models, types

EventHandler: TypeAlias = Callable[[models.Event], None]
HandlerTypeVar = TypeVar("HandlerTypeVar", bound=Callable[..., None])


class PGNoticeEventListener(asyncio.Queue[models.TableChangedEvent]):
    """Queue for PostgreSQL NOTIFY events."""


@dataclass
class EventRouter:
    """Dispatch parsed NOTIFY events to exactly one handler per `type`."""

    handlers: dict[types.EVENT_TYPES, EventHandler] = field(default_factory=dict, init=False)

    def register(self, event_type: types.EVENT_TYPES) -> Callable[[HandlerTypeVar], HandlerTypeVar]:
        """Decorator that wires *func* to *event_type*. The function must return `None`."""

        def decorator(func: HandlerTypeVar) -> HandlerTypeVar:
            if event_type in self.handlers:
                raise ValueError(f"duplicate handler for {event_type!r}")
            self.handlers[event_type] = func
            return func

        return decorator

    def __call__(self, envelope: models.AnyEvent) -> None:
        event = envelope.root
        try:
            self.handlers[event.type](event)
        except KeyError as exc:
            raise NotImplementedError(event) from exc


def default_event_router(
    *,
    notice_event_queue: PGNoticeEventListener,
    statistics: MutableMapping[str, models.EntrypointStatistics],
    canceled: MutableMapping[models.JobId, models.Context],
    pending_health_check: MutableMapping[uuid.UUID, asyncio.Future[models.HealthCheckEvent]],
) -> EventRouter:
    """Return an `EventRouter` wired with handlers for all known event types."""

    router = EventRouter()

    @router.register("table_changed_event")
    def _table_changed(evt: models.TableChangedEvent) -> None:
        notice_event_queue.put_nowait(evt)

    @router.register("requests_per_second_event")
    def _requests_per_second(evt: models.RequestsPerSecondEvent) -> None:  #
        for entrypoint, count in evt.entrypoint_count.items():
            statistics[entrypoint].samples.append((count, evt.sent_at))

    @router.register("cancellation_event")
    def _cancellation(evt: models.CancellationEvent) -> None:
        for jid in evt.ids:
            if ctx := canceled.get(jid):
                ctx.cancellation.cancel()

    @router.register("health_check_event")
    def _health_check_event(evt: models.HealthCheckEvent) -> None:
        if fut := pending_health_check.get(evt.id):
            fut.set_result(evt)

    return router


async def initialize_notice_event_listener(
    connection: db.Driver,
    channel: models.Channel,
    event_handler: Callable[[models.AnyEvent], None],
) -> None:
    """Add a listener on *channel* and funnel parsed events to *event_handler*."""

    def _process_payload(payload: str | bytes | bytearray) -> None:
        try:
            parsed = models.AnyEvent.model_validate_json(payload)
        except Exception as exc:
            logconfig.logger.critical(
                "Error parsing notification payload: %s", payload, exc_info=exc
            )
            return

        try:
            event_handler(parsed)
        except Exception as exc:
            logconfig.logger.critical(
                "Error while handling parsed event: %s", payload, exc_info=exc
            )

    await connection.add_listener(channel, _process_payload)



from __future__ import annotations

import asyncio
import signal
from contextlib import asynccontextmanager
from datetime import timedelta
from unittest.mock import ANY, MagicMock, patch

import async_timeout
import pytest

from pgqueuer import AsyncpgDriver, PgQueuer, QueueManager, SchedulerManager, supervisor
from pgqueuer.errors import FailingListenerError
from pgqueuer.models import HealthCheckEvent
from pgqueuer.types import QueueExecutionMode


@pytest.fixture(scope="function")
async def queue_manager(apgdriver: AsyncpgDriver) -> QueueManager:
    """Fixture to instantiate QueueManager."""
    return QueueManager(connection=apgdriver)


@pytest.fixture(scope="function")
async def scheduler_manager(apgdriver: AsyncpgDriver) -> SchedulerManager:
    """Fixture to instantiate SchedulerManager."""
    return SchedulerManager(connection=apgdriver)


@pytest.fixture(scope="function")
async def pg_queuer(apgdriver: AsyncpgDriver) -> PgQueuer:
    """Fixture to instantiate PgQueuer."""
    return PgQueuer(connection=apgdriver)


@pytest.fixture(scope="function")
def shutdown_event() -> asyncio.Event:
    """Fixture to create a shutdown asyncio.Event."""
    return asyncio.Event()


async def test_setup_shutdown_handlers_queue_manager(
    queue_manager: QueueManager,
    shutdown_event: asyncio.Event,
) -> None:
    manager = supervisor.setup_shutdown_handlers(queue_manager, shutdown_event)
    assert manager.shutdown is shutdown_event


async def test_setup_shutdown_handlers_scheduler_manager(
    scheduler_manager: SchedulerManager,
    shutdown_event: asyncio.Event,
) -> None:
    manager = supervisor.setup_shutdown_handlers(scheduler_manager, shutdown_event)
    assert manager.shutdown is shutdown_event


async def test_setup_shutdown_handlers_pg_queuer(
    pg_queuer: PgQueuer,
    shutdown_event: asyncio.Event,
) -> None:
    manager = supervisor.setup_shutdown_handlers(pg_queuer, shutdown_event)
    assert manager.shutdown is shutdown_event
    assert pg_queuer.qm.shutdown is shutdown_event
    assert pg_queuer.sm.shutdown is shutdown_event


async def test_setup_shutdown_handlers_invalid_manager(
    shutdown_event: asyncio.Event,
) -> None:
    class InvalidManager:
        pass

    with pytest.raises(NotImplementedError, match="Unsupported instance type: .*InvalidManager.*"):
        supervisor.setup_shutdown_handlers(InvalidManager(), shutdown_event)  # type: ignore[arg-type]


async def test_setup_signal_handlers(
    shutdown_event: asyncio.Event,
) -> None:
    with patch("pgqueuer.supervisor.asyncio.get_event_loop") as mock_get_event_loop:
        mock_loop = MagicMock()
        mock_get_event_loop.return_value = mock_loop

        supervisor.setup_signal_handlers(shutdown_event)

        mock_loop.add_signal_handler.assert_any_call(signal.SIGINT, ANY, signal.SIGINT)
        mock_loop.add_signal_handler.assert_any_call(signal.SIGTERM, ANY, signal.SIGTERM)


async def test_runit_normal_operation(
    pg_queuer: PgQueuer,
    shutdown_event: asyncio.Event,
) -> None:
    @asynccontextmanager
    async def foo():  # type: ignore
        yield pg_queuer

    # Run runit in the background
    runit_task = asyncio.create_task(
        supervisor.runit(
            factory=foo,
            dequeue_timeout=timedelta(seconds=1),
            batch_size=10,
            restart_delay=timedelta(seconds=2),
            restart_on_failure=False,
            shutdown=shutdown_event,
            mode=QueueExecutionMode.continuous,
            max_concurrent_tasks=None,
            shutdown_on_listener_failure=False,
        )
    )

    # Allow some time for runit to start
    await asyncio.sleep(0.1)

    # Signal shutdown
    shutdown_event.set()

    # Allow some time for runit to handle shutdown
    await asyncio.sleep(0.2)

    # Ensure runit_task is completed
    async with async_timeout.timeout(1):
        await runit_task
    assert runit_task.done()


async def test_runit_restart_on_failure(
    pg_queuer: PgQueuer,
    shutdown_event: asyncio.Event,
) -> None:
    calls = 0

    async def failing_run(*args, **kwargs):  # type: ignore
        nonlocal calls
        calls += 1
        raise Exception("Simulated failure")

    pg_queuer.run = failing_run  # type: ignore

    async def foo():  # type: ignore
        return pg_queuer

    # Run runit in the background
    runit_task = asyncio.create_task(
        supervisor.runit(
            factory=foo,
            dequeue_timeout=timedelta(seconds=1),
            batch_size=10,
            restart_delay=timedelta(seconds=0.05),
            restart_on_failure=True,
            shutdown=shutdown_event,
            mode=QueueExecutionMode.continuous,
            max_concurrent_tasks=None,
            shutdown_on_listener_failure=False,
        )
    )

    # Allow some time for runit to attempt running and failing
    await asyncio.sleep(0.2)

    # Signal shutdown to stop the runit loop
    shutdown_event.set()

    # Ensure runit_task is completed
    async with async_timeout.timeout(1):
        await runit_task
    assert runit_task.done()
    assert calls > 1


async def test_runit_no_restart_on_failure(
    pg_queuer: PgQueuer,
    shutdown_event: asyncio.Event,
) -> None:
    async def failing_run(*args, **kwargs):  # type: ignore
        raise Exception("Simulated failure")

    pg_queuer.run = failing_run  # type: ignore

    @asynccontextmanager
    async def foo():  # type: ignore
        yield pg_queuer

    with pytest.raises(Exception, match="Simulated failure"):
        await supervisor.runit(
            factory=foo,
            dequeue_timeout=timedelta(seconds=1),
            batch_size=10,
            restart_delay=timedelta(seconds=2),
            restart_on_failure=False,
            shutdown=shutdown_event,
            mode=QueueExecutionMode.continuous,
            max_concurrent_tasks=None,
            shutdown_on_listener_failure=False,
        )


async def test_runit_negative_restart_delay(shutdown_event: asyncio.Event) -> None:
    with pytest.raises(ValueError, match="'restart_delay' must be >= 0"):
        await supervisor.runit(
            factory=...,  # type: ignore
            dequeue_timeout=timedelta(seconds=1),
            batch_size=10,
            restart_delay=timedelta(seconds=-1),
            restart_on_failure=True,
            shutdown=shutdown_event,
            mode=QueueExecutionMode.continuous,
            max_concurrent_tasks=None,
            shutdown_on_listener_failure=False,
        )


async def test_run_manager_invalid_manager() -> None:
    class InvalidManager:
        pass

    with pytest.raises(NotImplementedError, match=r"Unsupported instance type: .*InvalidManager.*"):
        await supervisor.run_manager(
            InvalidManager(),  # type: ignore
            dequeue_timeout=timedelta(seconds=1),
            batch_size=10,
            mode=QueueExecutionMode.continuous,
            max_concurrent_tasks=None,
            shutdown_on_listener_failure=False,
        )


async def test_run_max_concurrent_tasks_twice_batch_size(
    queue_manager: QueueManager,
) -> None:
    with pytest.raises(
        RuntimeError,
        match=r"max_concurrent_tasks must be at least twice the batch size",
    ):
        await supervisor.run_manager(
            queue_manager,
            dequeue_timeout=timedelta(seconds=1),
            batch_size=10,
            mode=QueueExecutionMode.continuous,
            max_concurrent_tasks=10,
            shutdown_on_listener_failure=False,
        )


async def test_shutdown_on_listener_failure(queue_manager: QueueManager) -> None:
    with pytest.raises(FailingListenerError):
        await queue_manager.listener_healthy(timedelta(seconds=0))

    async def mocked_listener_healthy(timeout: timedelta) -> HealthCheckEvent:
        await asyncio.sleep(0.1)
        raise RuntimeError("Mocked")

    queue_manager.listener_healthy = mocked_listener_healthy  # type: ignore
    with pytest.raises(RuntimeError):
        await supervisor.run_manager(
            queue_manager,
            dequeue_timeout=timedelta(seconds=1),
            batch_size=10,
            mode=QueueExecutionMode.continuous,
            max_concurrent_tasks=None,
            shutdown_on_listener_failure=True,
        )



from __future__ import annotations

import asyncio
import dataclasses
from datetime import datetime, timedelta
from typing import Awaitable, Callable, Generic, TypeVar, cast

T = TypeVar("T")
UNSET = object()


@dataclasses.dataclass
class TTLCache(Generic[T]):
    """
    A Time-To-Live (TTL) cache that retains a value for a specified duration and
    updates it asynchronously upon expiration.

    This cache ensures that only one update occurs at a time. If the
    cached value is expired or has never been set, all callers will await the
    asynchronous update. If the update fails, the exception will propagate to every caller.

    Attributes:
        ttl: The duration that the cached value remains valid.
        on_expired: An asynchronous callable to refresh the value when it expires.
        expires_at: The datetime when the current cached value expires.
        value: The current cached value or a sentinel (UNSET) if not yet set.
        lock: An asyncio.Lock to prevent concurrent updates.
        update_task: A reference to the currently running update task, if any.
    """

    ttl: timedelta
    on_expired: Callable[[], Awaitable[T]]
    expires_at: datetime = dataclasses.field(init=False, default_factory=datetime.now)
    value: T | object = dataclasses.field(init=False, default=UNSET)
    lock: asyncio.Lock = dataclasses.field(init=False, default_factory=asyncio.Lock)
    update_task: asyncio.Task | None = dataclasses.field(init=False, default=None)

    async def __call__(self) -> T:
        """
        Retrieve the cached value. If the cached value is present and not
        expired, it is returned immediately.

        Otherwise, an asynchronous update is triggered and all callers await
        the update. If the update fails,its exception propagates to all callers.

        Returns:
            The cached value of type T.
        """
        now = datetime.now()
        if self.value is not UNSET and now <= self.expires_at:
            return cast(T, self.value)
        if self.update_task is None or self.update_task.done():
            self.expires_at = datetime.now() + self.ttl
            self.update_task = asyncio.create_task(self._update_value())
        await self.update_task
        return cast(T, self.value)

    async def _update_value(self) -> None:
        """
        Refresh the cached value asynchronously.

        This internal method is protected by an asyncio lock to ensure that only one
        update occurs at a time. It calls the on_expired callback to retrieve a new
        value, updates the cached value, and resets the expiration time.
        Exceptions from on_expired will propagate to the caller.
        """
        async with self.lock:
            self.value = await self.on_expired()

    @classmethod
    def create(cls, ttl: timedelta, on_expired: Callable[[], Awaitable[T]]) -> TTLCache[T]:
        """
        Create a new TTLCache instance with the specified time-to-live and asynchronous callback.

        Args:
            ttl: A timedelta representing how long the cached value remains valid.
            on_expired: An asynchronous function that provides a new value when the cache expires.

        Returns:
            A TTLCache instance configured with the provided ttl and on_expired callback.
        """
        return cls(ttl=ttl, on_expired=on_expired)



from .applications import PgQueuer
from .db import AsyncpgDriver, AsyncpgPoolDriver, PsycopgDriver
from .models import Job, JobId
from .qm import QueueManager
from .queries import Queries
from .sm import SchedulerManager

try:
    from ._version import version as __version__
except ImportError:
    __version__ = "0.0.0"  # Fallback version if _version.py is missing

__all__ = [
    "AsyncpgDriver",
    "AsyncpgPoolDriver",
    "Job",
    "JobId",
    "PgQueuer",
    "PsycopgDriver",
    "Queries",
    "QueueManager",
    "SchedulerManager",
]



from typing import cast

import psycopg
from flask import Flask, Response, g, jsonify, request

from pgqueuer.db import SyncPsycopgDriver, dsn
from pgqueuer.queries import SyncQueries


def create_app() -> Flask:
    app = Flask(__name__)

    def get_driver() -> SyncPsycopgDriver:
        if "driver" not in g:
            conn = psycopg.connect(dsn(), autocommit=True)
            g.driver = SyncPsycopgDriver(conn)

        return cast(SyncPsycopgDriver, g.driver)

    @app.teardown_appcontext
    def teardown_db(exception: BaseException | None) -> None:
        driver = cast(SyncPsycopgDriver | None, g.pop("driver", None))

        if driver is not None:
            driver._connection.close()

    @app.route("/enqueue", methods=["POST"])
    def enqueue() -> Response:
        queries = SyncQueries(get_driver())
        data = request.get_json(force=True)
        entrypoint = data.get("entrypoint", "default")
        payload = data.get("payload")
        priority = int(data.get("priority", 0))

        job_ids = queries.enqueue(entrypoint, payload.encode() if payload else None, priority)
        return jsonify({"job_ids": [str(jid) for jid in job_ids]})

    @app.route("/queue_size")
    def queue_size() -> Response:
        queries = SyncQueries(get_driver())
        stats = queries.queue_size()
        return jsonify(
            [
                {
                    "entrypoint": s.entrypoint,
                    "priority": s.priority,
                    "status": s.status,
                    "count": s.count,
                }
                for s in stats
            ]
        )

    return app


if __name__ == "__main__":
    create_app().run(debug=True)



from __future__ import annotations

import asyncio
import json
import os
import random
import signal
import sys
from contextlib import suppress
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from enum import Enum
from itertools import count
from pathlib import Path

import typer
import uvloop
from pydantic import AwareDatetime, BaseModel
from tabulate import tabulate
from tqdm.asyncio import tqdm

from pgqueuer import PgQueuer
from pgqueuer.db import AsyncpgDriver, AsyncpgPoolDriver, PsycopgDriver, dsn
from pgqueuer.models import Job
from pgqueuer.qb import add_prefix
from pgqueuer.queries import Queries

app = typer.Typer()


class DriverEnum(str, Enum):
    apg = "apg"
    apgpool = "apgpool"
    psy = "psy"


class BenchmarkResult(BaseModel):
    created_at: AwareDatetime
    driver: DriverEnum
    elapsed: timedelta
    github_ref_name: str
    queued: int
    rate: float
    steps: int

    def pretty_print(self) -> None:
        print(
            tabulate(
                [
                    ["Created At", self.created_at],
                    ["Driver", self.driver],
                    ["Elapsed Time", self.elapsed],
                    ["GitHub Ref Name", self.github_ref_name],
                    ["Rate", f"{self.rate:.2f}"],
                    ["Steps", self.steps],
                    ["Queued", self.queued],
                ],
                headers=["Field", "Value"],
                tablefmt=os.environ.get(add_prefix("TABLEFMT"), "pretty"),
                colalign=("left", "left"),
            )
        )


class Settings(BaseModel):
    driver: DriverEnum = typer.Option(
        DriverEnum.apg,
        help="Postgres driver to use.",
    )
    timer: timedelta = typer.Option(
        30.0,
        help="Run the benchmark for this number of seconds.",
    )
    dequeue: int = typer.Option(
        5,
        help="Number of concurrent dequeue tasks.",
    )
    dequeue_batch_size: int = typer.Option(
        10,
        help="Batch size for dequeue tasks.",
    )
    enqueue: int = typer.Option(
        1,
        help="Number of concurrent enqueue tasks.",
    )
    enqueue_batch_size: int = typer.Option(
        10,
        help="Batch size for enqueue tasks.",
    )
    requests_per_second: list[float] = typer.Option(
        [float("inf"), float("inf")],
        help="RPS for endpoints.",
    )
    concurrency_limit: list[int] = typer.Option(
        [sys.maxsize, sys.maxsize],
        help="Concurrency limit.",
    )
    output_json: Path | None = typer.Option(
        None,
        help="Output JSON file for benchmark metrics.",
    )

    def pretty_print(self) -> None:
        print(
            tabulate(
                [
                    ["Driver", self.driver],
                    ["Timer (s)", self.timer.total_seconds()],
                    ["Dequeue Tasks", self.dequeue],
                    ["Dequeue Batch Size", self.dequeue_batch_size],
                    ["Enqueue Tasks", self.enqueue],
                    ["Enqueue Batch Size", self.enqueue_batch_size],
                    ["Requests Per Second", self.requests_per_second],
                    ["Concurrency Limit", self.concurrency_limit],
                    ["Output JSON", self.output_json or "None"],
                ],
                headers=["Field", "Value"],
                tablefmt=os.environ.get(add_prefix("TABLEFMT"), "pretty"),
                colalign=("left", "left"),
            )
        )


async def make_queries(driver: DriverEnum, conninfo: str) -> Queries:
    match driver:
        case "apg":
            import asyncpg

            return Queries(AsyncpgDriver(await asyncpg.connect(dsn=conninfo)))
        case "apgpool":
            import asyncpg

            pool = await asyncpg.create_pool(dsn=conninfo)
            assert pool is not None
            return Queries(AsyncpgPoolDriver(pool))
        case "psy":
            import psycopg

            return Queries(
                PsycopgDriver(
                    await psycopg.AsyncConnection.connect(conninfo=conninfo, autocommit=True)
                )
            )
    raise NotImplementedError(driver)


@dataclass
class Consumer:
    pgq: PgQueuer
    batch_size: int
    entrypoint_rps: list[float]
    concurrency_limits: list[int]
    bar: tqdm

    async def run(self) -> None:
        async_rps, sync_rps = self.entrypoint_rps
        async_cl, sync_cl = self.concurrency_limits

        @self.pgq.entrypoint(
            "asyncfetch", requests_per_second=async_rps, concurrency_limit=async_cl
        )
        async def asyncfetch(job: Job) -> None:
            self.bar.update()

        @self.pgq.entrypoint("syncfetch", requests_per_second=sync_rps, concurrency_limit=sync_cl)
        def syncfetch(job: Job) -> None:
            self.bar.update()

        await self.pgq.run(batch_size=self.batch_size)


@dataclass
class Producer:
    shutdown: asyncio.Event
    queries: Queries
    batch_size: int
    cnt: count

    async def run(self) -> None:
        entrypoints = ["syncfetch", "asyncfetch"] * self.batch_size
        while not self.shutdown.is_set():
            await self.queries.enqueue(
                random.sample(entrypoints, k=self.batch_size),
                [f"{next(self.cnt)}".encode() for _ in range(self.batch_size)],
                [0] * self.batch_size,
            )


async def run_enqueuers(settings: Settings, shutdown: asyncio.Event) -> None:
    cnt = count()
    tasks = []
    for _ in range(settings.enqueue):
        q = await make_queries(settings.driver, dsn())
        p = Producer(shutdown, q, settings.enqueue_batch_size, cnt)
        tasks.append(p.run())
    await asyncio.gather(*tasks)


async def run_dequeuers(
    settings: Settings,
    pgqs: list[PgQueuer],
    tqdm_format_dict: dict,
) -> None:
    queries = [await make_queries(settings.driver, dsn()) for _ in range(settings.dequeue)]
    for q in queries:
        pgqs.append(PgQueuer(q.driver))
    with tqdm(ascii=True, unit=" job", unit_scale=True, file=sys.stdout) as bar:
        tasks = [
            Consumer(
                pgq,
                settings.dequeue_batch_size,
                settings.requests_per_second,
                settings.concurrency_limit,
                bar,
            ).run()
            for pgq in pgqs
        ]
        await asyncio.gather(*tasks)
        tqdm_format_dict.update(bar.format_dict)


async def benchmark(settings: Settings) -> None:
    settings.pretty_print()
    await (await make_queries(settings.driver, dsn())).clear_statistics_log()
    await (await make_queries(settings.driver, dsn())).clear_queue()

    shutdown = asyncio.Event()
    pgqs = list[PgQueuer]()
    tqdm_format_dict = dict[str, str]()

    async def shutdown_timer() -> None:
        with suppress(TimeoutError, asyncio.TimeoutError):
            await asyncio.wait_for(shutdown.wait(), settings.timer.total_seconds())
        shutdown.set()
        for q in pgqs:
            q.shutdown.set()

    def graceful_shutdown() -> None:
        shutdown.set()
        for qm in pgqs:
            qm.shutdown.set()

    loop = asyncio.get_event_loop()
    loop.add_signal_handler(signal.SIGINT, graceful_shutdown)
    loop.add_signal_handler(signal.SIGTERM, graceful_shutdown)

    await asyncio.gather(
        run_dequeuers(settings, pgqs, tqdm_format_dict),
        run_enqueuers(settings, shutdown),
        shutdown_timer(),
    )

    qsize = await (await make_queries(settings.driver, dsn())).queue_size()
    benchmark_result = BenchmarkResult(
        created_at=datetime.now(timezone.utc),
        driver=settings.driver,
        elapsed=tqdm_format_dict["elapsed"],
        github_ref_name=os.environ.get("REF_NAME", ""),
        queued=sum(x.count for x in qsize),
        rate=float(tqdm_format_dict["n"]) / float(tqdm_format_dict["elapsed"]),
        steps=tqdm_format_dict["n"],
    )
    if settings.output_json:
        with open(settings.output_json, "w") as f:
            json.dump(benchmark_result.model_dump(mode="json"), f)
    benchmark_result.pretty_print()


@app.command()
def main(
    driver: DriverEnum = typer.Option(DriverEnum.apg, "-d", "--driver"),
    timer: float = typer.Option(30.0, "-t", "--time"),
    dequeue: int = typer.Option(5),
    dequeue_batch_size: int = typer.Option(10),
    enqueue: int = typer.Option(1),
    enqueue_batch_size: int = typer.Option(10),
    requests_per_second: list[float] = typer.Option([float("inf"), float("inf")]),
    concurrency_limit: list[int] = typer.Option([0, 0]),
    output_json: Path | None = typer.Option(None),
) -> None:
    uvloop.run(
        benchmark(
            Settings(
                driver=driver,
                timer=timer,
                dequeue=dequeue,
                dequeue_batch_size=dequeue_batch_size,
                enqueue=enqueue,
                enqueue_batch_size=enqueue_batch_size,
                requests_per_second=requests_per_second,
                concurrency_limit=concurrency_limit,
                output_json=output_json,
            )
        )
    )


if __name__ == "__main__":
    with suppress(KeyboardInterrupt):
        app()



import asyncio
from typing import AsyncGenerator

import asyncpg
import psycopg
import pytest
import uvloop

from pgqueuer.db import AsyncpgDriver, SyncPsycopgDriver, dsn
from pgqueuer.queries import Queries


@pytest.fixture(scope="session", autouse=True)
def event_loop_policy() -> uvloop.EventLoopPolicy:
    return uvloop.EventLoopPolicy()


async def clear_all(driver: AsyncpgDriver) -> None:
    await asyncio.gather(
        Queries(driver).clear_queue_log(),
        Queries(driver).clear_queue(),
        Queries(driver).clear_schedule(),
        Queries(driver).clear_statistics_log(),
    )


@pytest.fixture(scope="function")
async def apgdriver() -> AsyncGenerator[AsyncpgDriver, None]:
    conn = await asyncpg.connect(dsn=dsn())
    driver = AsyncpgDriver(conn)
    await clear_all(driver)
    try:
        yield driver
    finally:
        await conn.close()


@pytest.fixture(scope="function")
async def pgdriver(apgdriver: AsyncpgDriver) -> AsyncGenerator[SyncPsycopgDriver, None]:
    conn = psycopg.connect(dsn(), autocommit=True)
    try:
        yield SyncPsycopgDriver(conn)
    finally:
        conn.close()



from __future__ import annotations


class PgqException(Exception):
    """Base class for all exceptions raised by PGQueuer."""


class RetryException(PgqException):
    """Exception raised for retry-related errors in PGQueuer."""


class MaxRetriesExceeded(RetryException):
    """Exception raised when all retry attempts have been exhausted."""


class MaxTimeExceeded(RetryException):
    """Exception raised when the maximum time limit for retries has been exceeded."""


class DuplicateJobError(PgqException):
    """Raised when enqueue violates a deduplication constraint."""

    def __init__(self, dedupe_key: list[str | None]) -> None:
        super().__init__()
        self.dedupe_key = dedupe_key


class FailingListenerError(PgqException):
    """Raised when a listener fails to process a job."""



from __future__ import annotations

import time
from datetime import datetime, timedelta

import pytest

from pgqueuer.helpers import (
    ExponentialBackoff,
    add_schema_to_dsn,
    normalize_cron_expression,
    retry_timer_buffer_timeout,
    timeout_with_jitter,
    timer,
    utc_now,
)


async def test_perf_counter_dt() -> None:
    assert isinstance(utc_now(), datetime)
    assert utc_now().tzinfo is not None


def test_heartbeat_buffer_timeout_empty_list() -> None:
    dts = list[timedelta]()
    expected = timedelta(hours=24)
    assert retry_timer_buffer_timeout(dts) == expected


def test_heartbeat_buffer_timeout_all_dts_less_than_or_equal_to_t0() -> None:
    dts = [timedelta(seconds=-1), timedelta(seconds=0)]
    expected = timedelta(hours=24)
    assert retry_timer_buffer_timeout(dts) == expected


def test_heartbeat_buffer_timeout_positive_dts() -> None:
    dts = [timedelta(seconds=10), timedelta(seconds=5)]
    expected = timedelta(seconds=5)
    assert retry_timer_buffer_timeout(dts) == expected


def test_heartbeat_buffer_timeout_mixed_dts() -> None:
    dts = [timedelta(seconds=-5), timedelta(seconds=10)]
    expected = timedelta(seconds=10)
    assert retry_timer_buffer_timeout(dts) == expected


def test_heartbeat_buffer_timeout_custom_t0() -> None:
    dts = [timedelta(seconds=4), timedelta(seconds=6)]
    expected = timedelta(seconds=6)
    assert retry_timer_buffer_timeout(dts, _t0=timedelta(seconds=5)) == expected


def test_heartbeat_buffer_timeout_custom_default() -> None:
    dts = list[timedelta]()
    expected = timedelta(hours=48)
    assert retry_timer_buffer_timeout(dts, _default=timedelta(hours=48)) == expected


def test_delay_within_jitter_range() -> None:
    base_timeout = timedelta(seconds=10)
    jitter_span = (0.8, 1.2)

    # Call the function multiple times to check the jitter range
    for _ in range(100):
        delay = timeout_with_jitter(base_timeout, jitter_span)
        base_delay = base_timeout.total_seconds()
        assert base_delay * jitter_span[0] <= delay.total_seconds() <= base_delay * jitter_span[1]


def test_delay_is_timedelta() -> None:
    base_timeout = timedelta(seconds=5)
    delay = timeout_with_jitter(base_timeout)
    assert isinstance(delay, timedelta)


def test_custom_jitter_range() -> None:
    base_timeout = timedelta(seconds=8)
    jitter_span = (0.5, 1.5)

    # Call the function multiple times to check the custom jitter range
    for _ in range(100):
        delay = timeout_with_jitter(base_timeout, jitter_span)
        base_delay = base_timeout.total_seconds()
        assert base_delay * jitter_span[0] <= delay.total_seconds() <= base_delay * jitter_span[1]


@pytest.mark.parametrize(
    "expression, expected",
    (
        ("@hourly", "0 * * * *"),
        ("@midnight", "0 0 * * *"),
    ),
)
def test_normalize_cron_expression(expression: str, expected: str) -> None:
    assert normalize_cron_expression(expression) == expected


def test_timer() -> None:
    with timer() as elapsed:
        t1 = elapsed()
        time.sleep(0.01)
        t2 = elapsed()
        assert t2 > t1

    assert elapsed() == elapsed()

    with pytest.raises(ValueError):
        with timer() as elapsed:
            raise ValueError

        assert elapsed() == elapsed()


def test_add_schema_to_empty_dsn() -> None:
    dsn = "postgresql://user:password@host:port/dbname"
    schema = "myschema"
    expected = "postgresql://user:password@host:port/dbname?options=-csearch_path%3Dmyschema"
    assert add_schema_to_dsn(dsn, schema) == expected


def test_add_schema_to_dsn_with_existing_query() -> None:
    dsn = "postgresql://user:password@host:port/dbname?sslmode=require"
    schema = "myschema"
    expected = "postgresql://user:password@host:port/dbname?sslmode=require&options=-csearch_path%3Dmyschema"
    assert add_schema_to_dsn(dsn, schema) == expected


def test_raise_on_existing_search_path() -> None:
    dsn = "postgresql://user:password@host:port/dbname?options=-c+search_path=otherschema"
    schema = "myschema"
    with pytest.raises(ValueError, match="search_path is already set in the options parameter."):
        add_schema_to_dsn(dsn, schema)


def test_preserve_other_options_and_add_search_path() -> None:
    dsn = "postgresql://user:password@host:port/dbname?options=-cother_option=foo"
    schema = "myschema"
    expected = "postgresql://user:password@host:port/dbname?options=-cother_option%3Dfoo&options=-csearch_path%3Dmyschema"
    assert add_schema_to_dsn(dsn, schema) == expected


def test_exponential_backoff_initial_delay() -> None:
    backoff = ExponentialBackoff(
        start_delay=timedelta(1),
        multiplier=2,
        max_delay=timedelta(10),
    )
    assert backoff.current_delay == timedelta(1)


def test_exponential_backoff_next_delay() -> None:
    backoff = ExponentialBackoff(
        start_delay=timedelta(1),
        multiplier=2,
        max_delay=timedelta(10),
    )
    assert backoff.next_delay() == timedelta(2)
    assert backoff.next_delay() == timedelta(4)
    assert backoff.next_delay() == timedelta(8)
    assert backoff.next_delay() == timedelta(10)  # Capped at max_limit
    assert backoff.next_delay() == timedelta(10)  # Capped at max_limit


def test_exponential_backoff_max_limit() -> None:
    backoff = ExponentialBackoff(
        start_delay=timedelta(3),
        multiplier=3,
        max_delay=timedelta(20),
    )
    backoff.next_delay()
    assert backoff.current_delay == timedelta(9)
    backoff.next_delay()
    assert backoff.current_delay == timedelta(20)  # Capped at max_limit
    assert backoff.current_delay == timedelta(20)  # Capped at max_limit


def test_exponential_backoff_reset() -> None:
    backoff = ExponentialBackoff(
        start_delay=timedelta(5),
        multiplier=2,
        max_delay=timedelta(50),
    )
    backoff.next_delay()
    backoff.next_delay()
    assert backoff.current_delay == timedelta(20)
    backoff.reset()
    assert backoff.current_delay == timedelta(5)
